xquery version "3.1";

(:~
 : This module is a refactored version of the old presort module which lead to
 : unidentifiable errors while transforming the original encoding to another TEI.
 :
 : It mainly deals with bringing everything into the right order, i.e. in a fully
 : legible flow of text that has the right chronological flow.
 :
 :)

module namespace fsort="http://fontane-nb.dariah.eu/sort";

declare namespace tei="http://www.tei-c.org/ns/1.0";

import module namespace config="http://textgrid.de/ns/SADE/config" at "../../config/config.xqm";
import module namespace presort="http://fontane-nb.dariah.eu/presort" at "presort.xqm";

(:~
 : The main function. Sorts each notebook and saves a sorted version at
 : /db/sade-projects/textgrid/data/xml/print/xml/notebook_uri-presort.xml.
 :
 : @author Michelle Weidling
 :)
declare function fsort:main($tei as node()*, $log as xs:string) as element(tei:TEI) {
    let $prepared := presort:prepare($tei)
    let $prepared := fsort:enhance-handshifts($prepared, $log)
    let $tei := fsort:sort($prepared, $log)
    let $fully-sorted := presort:sort-integrations($tei)
    let $id := $tei/@id

    let $store := xmldb:store($config:data-root || "/print/xml/", $id || "-presort.xml", $fully-sorted)
    return $fully-sorted
};


(:~
 : The main sorting routine. Everything apart from tei:hi and tei:rs follows the
 : same rules that are specified in fsort:default-return().
 :
 :)
declare function fsort:sort($nodes as node()*, $log as xs:string) {
    for $node in $nodes return
        typeswitch ( $node )
        case text() return
            $node

        case comment() return
            if(matches($node, "transform")) then
                $node
            else
                ()

        case element(tei:zone) return
            fsort:default-return($node, $log)

        case element(tei:line) return
            fsort:default-return($node, $log)

        case element(tei:seg) return
            fsort:default-return($node, $log)

        case element(tei:rs) return
            fsort:default-return($node, $log)

(:        case element(tei:hi) return:)
(:            if($node/@next and not($node/ancestor::tei:rs)) then:)
(:                let $corresp-node := fsort:find-corresp-node($node, "next"):)
(:                return:)
(:                    (: it's not sufficient to copy only the corresp node since:)
(:                    this would mean we'd loose the information about the tei:rs :):)
(:                    if($corresp-node[self::tei:hi and parent::tei:rs]) then:)
(:                        (fsort:keep-node($node),:)
(:                        element {QName("http://www.tei-c.org/ns/1.0", "rs")} {:)
(:                            $corresp-node/parent::*/@*,:)
(:                            $corresp-node:)
(:                        }):)
(:                    else:)
(:                        fsort:default-return($node):)
(:            else:)
(:                fsort:default-return($node):)
(::)
(::)
(:        case element(tei:rs) return:)
(:            (: in this case the tei:rs is empty after fsort:apply-all-nexts. we:)
(:            don't need these. :):)
(:            if($node[child::*[@prev]] and count($node/child::*) = 1) then:)
(:                ():)
(:            else:)
(:                fsort:default-return($node):)
(:                :)
(:        case element(tei:date) return:)
(:            fsort:keep-node($node):)
(::)
(:        case element(tei:ref) return:)
(:            fsort:keep-node($node):)

        default return
            fsort:keep-node($node, "sort", $log)
};

(:~
 : Handles what happens with a node. We distinguish between nodes that are at the
 : start of a virtual aggregation, in its middle, its end, and interlinear
 : additions.
 :
 : @author Michelle Weidling
 : @param $node the current TEI element
 : @return a copy of the current element, a fully sorted virtual aggregation or an empty sequence
 :)
declare function fsort:default-return($node as node(), $log as xs:string) as node()* {
    (: nodes with @next and @prev are in the middle of a virtual aggregation and
    handled in fsort:apply-all-nexts, thus we only look at the first element of
    the aggregation.
    furthermore, when nodes that indicate an underlined section of text (@style = "text-decoration: underline"
    or similar), their @prev/@next doesn't indicate the chronology but only the
    continuation of the highlighting. they should thus be omitted.:)
    if(matches($node/@style, "underline")) then
        fsort:keep-node($node, "sort", $log)

    else if($node[@next and not(@prev)]) then
        fsort:apply-all-nexts($node, $log)

    (: since all parts of the virtual aggregation are handled by fsort:apply-all-nexts
    we can ignore the ones that have a @prev :)
    else if($node[@prev]) then
        ()

    (: we distinguish this case and the one below to improve performance :)
    else if($node[descendant::*[@next or @prev]]) then
        fsort:keep-node($node, "sort", $log)

    (: nodes with this attribute are part of an interlinear addition; they have
    been marked by presort:prepare. since interlinear additions are handled as a
    whole and put into the right place, their elements can be safely dumped here :)
    else if($node[@type = "interlinear"]) then
        ()

    else
        $node
};


(:
 : This functions establishes the correct chronology of elements which is
 : indicated by a virtual aggregation (@prev/@next). It's called by the first
 : node that has a @next (but not a @prev) and recursively finds the next
 : node in the aggregation.
 :
 : @author Michelle Weidling
 : @param $node the current node with a @next and/or @prev attribute
 : @return a sequence of all nodes belonging to the virtual aggregation in the
 : right order
 :  :)
declare function fsort:apply-all-nexts($node as node(), $log as xs:string) as node()* {
    (: in case an addSpan is part of a virtual aggregation :)
    if($node[self::tei:addSpan]) then
        let $spanTo := substring-after($node/@spanTo, "#")
        let $anchor := $node/following::tei:anchor[@xml:id = $spanTo]
        let $nodes-inbetween := $node/following-sibling::*[. << $anchor]

        return
            for $node-inbetween in $nodes-inbetween return
                fsort:keep-node($node-inbetween, "sort", $log)

    else
        (: first element of a virtual aggregation: entry point :)
        if($node/@next and not($node/@prev)) then
            let $next-node := fsort:find-corresp-node($node, "next")
            return
                if(count($next-node) = 1) then
                    (fsort:keep-node($node, "sort", $log),
                    (: check if the parts are in different lines :)
                    if($node/ancestor::tei:line = $next-node/ancestor::tei:line) then
                        ()
                    else
                        element tei:milestone {
                            attribute unit {"line"}
                        },
                    fsort:apply-all-nexts($next-node, $log))
                else if(not($next-node)) then
                    fsort:add-log-entry($log, "No next node found for " || $node/@next)
                else
                    fsort:add-log-entry($log, "Several next nodes found for " || $node/@next)

        (: last of a virtual aggregation: exit point :)
        else if(not($node/@next)) then
            ($node/preceding::tei:handShift[1],
            fsort:keep-node($node, "sort", $log))

        (: element in the middle of a virtual aggregation:)
        else
            let $next-node := fsort:find-corresp-node($node, "next")
            let $prev-handshift := $node/preceding::tei:handShift[1]
            return
                if(count($next-node) = 1) then
                    ($prev-handshift,
                    fsort:keep-node($node, "sort", $log),
                    (: check if the parts are in different lines :)
                    if($node/ancestor::tei:line = $next-node/ancestor::tei:line) then
                        ()
                    else
                        element tei:milestone {
                            attribute unit {"line"}
                        },
                    fsort:apply-all-nexts($next-node, $log))
                else if(not($next-node)) then
                    fsort:add-log-entry($log, "No next node found for " || $node/@next)
                else
                    fsort:add-log-entry($log, "Several next nodes found for " || $node/@next)
};


(:~
 : In cases where a node has a descendant one with a @prev or @next attribute
 : the current node has to be kept, but its descendants have to be sorted.
 :
 : In cases where a node DOESN'T have a descendant with said attributes, the
 : whole node is kept in fsort:default-return.
 :
 : This procedure should improve perfomance when dealing with all notebooks
 : (which takes some time).
 :
 : @author Michelle Weidling
 : @param $node the current node
 : @return a copy of the current node with sorted descendants
 :)
declare function fsort:keep-node($node as node(), $flag as xs:string, $log as xs:string) as node() {
    element {QName("http://www.tei-c.org/ns/1.0", $node/name())} {
        $node/@*,
        if($flag = "sort") then
            fsort:sort($node/node(), $log)
        else
            fsort:enhance-handshifts($node/node(), $log)
    }
};



declare function fsort:find-corresp-node($node as node(), $flag as xs:string) {
    let $target-id :=
        if($flag = "next") then
            fsort:get-id($node/@next)
        else if($flag = "prev") then
            fsort:get-id($node/@prev)
        else
            error(QName("FONTANE", "fsort1"), "Invalid flag: " || $flag || "." )
    return
        $node/ancestor::*[last()]//*[@xml:id = $target-id]
};


declare function fsort:get-id($target as xs:string) as xs:string {
    $target => replace("#", "")
};


(:~
 : tei:rs oftentimes has one or several element(s) as descendant(s) which can be in
 : the middle or end of a virtual aggregation. These have already been handled by
 : sort:apply-all-nexts and should therefore not be considered.
 :
 : @author Michelle Weidling
 : @param $node the current tei:rs
 : @return tei:rs
 :)
declare function fsort:exclude-copied($node as element(tei:rs)) as
element(tei:rs)? {
    if(fsort:has-only-copied-children($node)) then
        ()
    else
        let $rs-children := $node/node()
        let $processed-children :=
            for $child in $rs-children return
                if($child/@next
                and not(matches($child/@style, "underline"))) then
                    fsort:apply-all-nexts($child, $log)
                else
                    $child

        return
            if(count($processed-children) gt 0) then
                element {QName("http://www.tei-c.org/ns/1.0", $node/name())} {
                    $node/@*,
                    $processed-children
                }
            else
                ()
};

(:~
 : tei:rs oftentimes has one or several element(s) as descendant(s) which can be part
 : of a virtual aggregation. If this applies to all children of tei:rs, all of them
 : have been handled by fsort:apply-all-nexts, thus leaving the tei:rs empty.
 :
 : In these cases we need an indicator if we can savely drop a tei:rs.
 :
 : @author Michelle Weidling
 : @param $node the current tei:rs
 : @return xs:boolean
 :  :)
declare function fsort:has-only-copied-children($node as element()) as
xs:boolean {
    let $are-copied :=
        for $desc in $node//node() return
            if($desc[@prev]) then
                true()
            else
                false()
    return
        not($are-copied = false())
};


declare function fsort:add-log-entry($log-file as xs:string,
$message as xs:string) as empty-sequence() {
  let $entry := <LogEntry timestamp="{util:system-time()}">{$message}</LogEntry>
  return update insert $entry into doc($log-file)/*
};



(:~
 : Adds the information back to tei:handShifts that have been omitted in order to
 : avoid redundancy.
 :
 :)
declare function fsort:enhance-handshifts($nodes as node()*, $log as xs:string)
as node()* {
    for $node in $nodes return
        typeswitch ($node)

        case text() return
            $node

        case comment() return
            if(matches($node, "transform")) then
                $node
            else
                ()

        case element(tei:handShift) return
            element tei:handShift {
                if($node/@new) then
                    $node/@new
                else
                    attribute new {$node/preceding::tei:handShift[@new][1]/@new},

                if($node/@medium) then
                    $node/@medium
                else
                    attribute medium {$node/preceding::tei:handShift[@medium][1]/@medium},

                if($node/@script) then
                    $node/@script
                else
                    attribute script {$node/preceding::tei:handShift[@script][1]/@script}
            }

        default return
            fsort:keep-node($node, "handshift", $log)
};
